@@ -1,7 +1,7 @@
 -- |
 -- Module:      Data.Aeson.Internal
--- Copyright:   (c) 2015 Bryan O'Sullivan
--- License:     Apache
+-- Copyright:   (c) 2015-2016 Bryan O'Sullivan
+-- License:     BSD3
 -- Maintainer:  Bryan O'Sullivan <bos@serpentine.com>
 -- Stability:   experimental
 -- Portability: portable
@@ -14,8 +14,9 @@
 module Data.Aeson.Internal
     (
       IResult(..)
-    , JSONPathElement
+    , JSONPathElement(..)
     , JSONPath
+    , (<?>)
     , formatError
     , ifromJSON
     , iparse
