@@ -1,8 +1,8 @@
 -- |
 -- Module:      Data.Aeson.Types
--- Copyright:   (c) 2011-2015 Bryan O'Sullivan
+-- Copyright:   (c) 2011-2016 Bryan O'Sullivan
 --              (c) 2011 MailRank, Inc.
--- License:     Apache
+-- License:     BSD3
 -- Maintainer:  Bryan O'Sullivan <bos@serpentine.com>
 -- Stability:   experimental
 -- Portability: portable
@@ -14,6 +14,7 @@ module Data.Aeson.Types
     -- * Core JSON types
       Value(..)
     , Encoding
+    , unsafeToEncoding
     , fromEncoding
     , Series
     , Array
@@ -39,6 +40,7 @@ module Data.Aeson.Types
     -- ** Generic JSON classes
     , GFromJSON(..)
     , GToJSON(..)
+    , GToEncoding(..)
     , genericToJSON
     , genericToEncoding
     , genericParseJSON
@@ -55,6 +57,7 @@ module Data.Aeson.Types
     , foldable
     , (.:)
     , (.:?)
+    , (.:!)
     , (.!=)
     , object
 
