@@ -1,8 +1,17 @@
 {-# LANGUAGE CPP, BangPatterns, DeriveDataTypeable, FlexibleContexts,
-    FlexibleInstances, GeneralizedNewtypeDeriving, IncoherentInstances,
-    OverlappingInstances, OverloadedStrings, UndecidableInstances,
+    FlexibleInstances, GeneralizedNewtypeDeriving,
+    OverloadedStrings, UndecidableInstances,
     ViewPatterns #-}
 {-# LANGUAGE DefaultSignatures #-}
+
+-- Needed for Tagged, Const and Proxy instances
+#if __GLASGOW_HASKELL__ >= 706
+{-# LANGUAGE PolyKinds #-}
+#endif
+
+#define NEEDS_INCOHERENT
+#include "overlapping-compat.h"
+
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
 -- TODO: Drop this when we remove support for Data.Attoparsec.Number
@@ -10,9 +19,9 @@
 
 -- |
 -- Module:      Data.Aeson.Types.Instances
--- Copyright:   (c) 2011-2015 Bryan O'Sullivan
+-- Copyright:   (c) 2011-2016 Bryan O'Sullivan
 --              (c) 2011 MailRank, Inc.
--- License:     Apache
+-- License:     BSD3
 -- Maintainer:  Bryan O'Sullivan <bos@serpentine.com>
 -- Stability:   experimental
 -- Portability: portable
@@ -29,6 +38,7 @@ module Data.Aeson.Types.Instances
     -- ** Generic JSON classes
     , GFromJSON(..)
     , GToJSON(..)
+    , GToEncoding(..)
     , genericToJSON
     , genericToEncoding
     , genericParseJSON
@@ -49,35 +59,43 @@ module Data.Aeson.Types.Instances
     , ifromJSON
     , (.:)
     , (.:?)
+    , (.:!)
     , (.!=)
     , tuple
     , (>*<)
     , typeMismatch
     ) where
 
-import Control.Applicative ((<$>), (<*>), pure)
+import Control.Applicative (Const(..))
 import Data.Aeson.Encode.Functions (brackets, builder, encode, foldable, list)
 import Data.Aeson.Functions (hashMapKey, mapHashKeyVal, mapKey, mapKeyVal)
 import Data.Aeson.Types.Class
 import Data.Aeson.Types.Internal
 import Data.Attoparsec.Number (Number(..))
 import Data.Fixed (Fixed, HasResolution)
-import Data.Foldable (Foldable, toList)
+import Data.Foldable (toList)
 import Data.Functor.Identity (Identity(..))
 import Data.Hashable (Hashable(..))
 import Data.Int (Int8, Int16, Int32, Int64)
+import Data.List.NonEmpty (NonEmpty(..))
 import Data.Maybe (fromMaybe)
-import Data.Monoid ((<>), mempty)
+import Data.Monoid ((<>))
 import Data.Monoid (Dual(..), First(..), Last(..))
+import Data.Proxy (Proxy(..))
 import Data.Ratio (Ratio, (%), numerator, denominator)
 import Data.Scientific (Scientific)
+import Data.Tagged (Tagged(..))
 import Data.Text (Text, pack, unpack)
-import Data.Time (Day, LocalTime, NominalDiffTime, UTCTime, ZonedTime)
+import Data.Time (Day, LocalTime, NominalDiffTime, TimeOfDay, UTCTime,
+                  ZonedTime)
 import Data.Time.Format (FormatTime, formatTime, parseTime)
-import Data.Traversable as Tr (sequence, traverse)
+import Data.Traversable as Tr (sequence)
 import Data.Vector (Vector)
-import Data.Word (Word, Word8, Word16, Word32, Word64)
+import Data.Word (Word8, Word16, Word32, Word64)
+import Data.Version (Version, showVersion, parseVersion)
+import Text.ParserCombinators.ReadP (readP_to_S)
 import Foreign.Storable (Storable)
+import Numeric.Natural (Natural)
 import Prelude hiding (foldr)
 import qualified Data.Aeson.Encode.Builder as E
 import qualified Data.Aeson.Parser.Time as Time
@@ -102,6 +120,13 @@ import qualified Data.Vector.Primitive as VP
 import qualified Data.Vector.Storable as VS
 import qualified Data.Vector.Unboxed as VU
 
+#if MIN_VERSION_base(4,8,0)
+#else
+import Control.Applicative ((<$>), (<*>), pure)
+import Data.Monoid (mempty)
+import Data.Word (Word)
+#endif
+
 #if MIN_VERSION_time(1,5,0)
 import Data.Time.Format (defaultTimeLocale)
 #else
@@ -142,15 +167,15 @@ instance (ToJSON a, ToJSON b) => ToJSON (Either a b) where
     {-# INLINE toJSON #-}
 
     toEncoding (Left a) = Encoding $
-      B.shortByteString "{\"left\":" <> builder a <> B.char7 '}'
+      B.shortByteString "{\"Left\":" <> builder a <> B.char7 '}'
     toEncoding (Right a) = Encoding $
-      B.shortByteString "{\"right\":" <> builder a <> B.char7 '}'
+      B.shortByteString "{\"Right\":" <> builder a <> B.char7 '}'
     {-# INLINE toEncoding #-}
 
 instance (FromJSON a, FromJSON b) => FromJSON (Either a b) where
     parseJSON (Object (H.toList -> [(key, value)]))
-        | key == left  = Left  <$> parseJSON value
-        | key == right = Right <$> parseJSON value
+        | key == left  = Left  <$> parseJSON value <?> Key left
+        | key == right = Right <$> parseJSON value <?> Key right
     parseJSON _        = fail $
         "expected an object with a single property " ++
         "where the property key should be either " ++
@@ -172,6 +197,24 @@ instance FromJSON Bool where
     parseJSON = withBool "Bool" pure
     {-# INLINE parseJSON #-}
 
+instance ToJSON Ordering where
+  toJSON     = toJSON     . orderingToText
+  toEncoding = toEncoding . orderingToText
+
+orderingToText :: Ordering -> T.Text
+orderingToText o = case o of
+                     LT -> "LT"
+                     EQ -> "EQ"
+                     GT -> "GT"
+
+instance FromJSON Ordering where
+  parseJSON = withText "Ordering" $ \s ->
+    case s of
+      "LT" -> return LT
+      "EQ" -> return EQ
+      "GT" -> return GT
+      _ -> fail "Parsing Ordering value failed: expected \"LT\", \"EQ\", or \"GT\""
+
 instance ToJSON () where
     toJSON _ = emptyArray
     {-# INLINE toJSON #-}
@@ -186,14 +229,14 @@ instance FromJSON () where
                     else fail "Expected an empty array"
     {-# INLINE parseJSON #-}
 
-instance ToJSON [Char] where
+instance INCOHERENT_ ToJSON [Char] where
     toJSON = String . T.pack
     {-# INLINE toJSON #-}
 
     toEncoding = Encoding . E.string
     {-# INLINE toEncoding #-}
 
-instance FromJSON [Char] where
+instance INCOHERENT_ FromJSON [Char] where
     parseJSON = withText "String" $ pure . T.unpack
     {-# INLINE parseJSON #-}
 
@@ -318,6 +361,19 @@ instance FromJSON Integer where
     parseJSON = withScientific "Integral" $ pure . truncate
     {-# INLINE parseJSON #-}
 
+instance ToJSON Natural where
+    toJSON = toJSON . toInteger
+    {-# INLINE toJSON #-}
+
+    toEncoding = toEncoding . toInteger
+    {-# INLINE toEncoding #-}
+
+instance FromJSON Natural where
+    parseJSON = withScientific "Natural" $ \s ->
+      if Scientific.coefficient s < 0
+        then fail $ "Expected a Natural number but got the negative number: " <> show s
+        else pure $ truncate s
+
 instance ToJSON Int8 where
     toJSON = Number . fromIntegral
     {-# INLINE toJSON #-}
@@ -441,19 +497,33 @@ instance FromJSON LT.Text where
     parseJSON = withText "Lazy Text" $ pure . LT.fromStrict
     {-# INLINE parseJSON #-}
 
-instance (ToJSON a) => ToJSON [a] where
+instance (ToJSON a) => ToJSON (NonEmpty a) where
+    toJSON = toJSON . toList
+    {-# INLINE toJSON #-}
+
+    toEncoding = toEncoding . toList
+    {-# INLINE toEncoding #-}
+
+instance (FromJSON a) => FromJSON (NonEmpty a) where
+    parseJSON = withArray "NonEmpty a" $
+        (>>= ne) . Tr.sequence . zipWith parseIndexedJSON [0..] . V.toList
+      where
+        ne []     = fail "Expected a NonEmpty but got an empty list"
+        ne (x:xs) = pure (x :| xs)
+
+instance OVERLAPPABLE_ (ToJSON a) => ToJSON [a] where
     toJSON = Array . V.fromList . map toJSON
     {-# INLINE toJSON #-}
 
     toEncoding xs = list xs
     {-# INLINE toEncoding #-}
 
-instance (FromJSON a) => FromJSON [a] where
+instance OVERLAPPABLE_ (FromJSON a) => FromJSON [a] where
     parseJSON = withArray "[a]" $ Tr.sequence .
                 zipWith parseIndexedJSON [0..] . V.toList
     {-# INLINE parseJSON #-}
 
-instance (Foldable t, ToJSON a) => ToJSON (t a) where
+instance (ToJSON a) => ToJSON (Seq.Seq a) where
     toJSON = toJSON . toList
     {-# INLINE toJSON #-}
 
@@ -461,7 +531,9 @@ instance (Foldable t, ToJSON a) => ToJSON (t a) where
     {-# INLINE toEncoding #-}
 
 instance (FromJSON a) => FromJSON (Seq.Seq a) where
-    parseJSON = withArray "Seq a" $ traverse parseJSON . Seq.fromList . V.toList
+    parseJSON = withArray "Seq a" $
+      fmap Seq.fromList .
+      Tr.sequence . zipWith parseIndexedJSON [0..] . V.toList
     {-# INLINE parseJSON #-}
 
 instance (ToJSON a) => ToJSON (Vector a) where
@@ -713,14 +785,21 @@ instance FromJSON DotNetTime where
 
 instance ToJSON Day where
     toJSON       = stringEncoding
-    toEncoding z = Encoding (E.day z)
+    toEncoding z = Encoding (E.quote $ E.day z)
 
 instance FromJSON Day where
     parseJSON = withText "Day" (Time.run Time.day)
 
+instance ToJSON TimeOfDay where
+    toJSON       = stringEncoding
+    toEncoding z = Encoding (E.quote $ E.timeOfDay z)
+
+instance FromJSON TimeOfDay where
+    parseJSON = withText "TimeOfDay" (Time.run Time.timeOfDay)
+
 instance ToJSON LocalTime where
     toJSON       = stringEncoding
-    toEncoding z = Encoding (E.localTime z)
+    toEncoding z = Encoding (E.quote $ E.localTime z)
 
 instance FromJSON LocalTime where
     parseJSON = withText "LocalTime" (Time.run Time.localTime)
@@ -728,7 +807,7 @@ instance FromJSON LocalTime where
 instance ToJSON ZonedTime where
     toJSON = stringEncoding
 
-    toEncoding z = Encoding (E.zonedTime z)
+    toEncoding z = Encoding (E.quote $ E.zonedTime z)
 
 formatMillis :: (FormatTime t) => t -> String
 formatMillis = take 3 . formatTime defaultTimeLocale "%q"
@@ -739,11 +818,11 @@ instance FromJSON ZonedTime where
 instance ToJSON UTCTime where
     toJSON = stringEncoding
 
-    toEncoding t = Encoding (E.utcTime t)
+    toEncoding t = Encoding (E.quote $ E.utcTime t)
 
 -- | Encode something to a JSON string.
 stringEncoding :: (ToJSON a) => a -> Value
-stringEncoding = String . T.decodeLatin1 . L.toStrict . encode
+stringEncoding = String . T.dropAround (== '"') . T.decodeLatin1 . L.toStrict . encode
 {-# INLINE stringEncoding #-}
 
 instance FromJSON UTCTime where
@@ -1456,6 +1535,55 @@ instance FromJSON a => FromJSON (Last a) where
     parseJSON = fmap Last . parseJSON
     {-# INLINE parseJSON #-}
 
+instance ToJSON Version where
+    toJSON = toJSON . showVersion
+    {-# INLINE toJSON #-}
+
+    toEncoding = toEncoding . showVersion
+    {-# INLINE toEncoding #-}
+
+instance FromJSON Version where
+    {-# INLINE parseJSON #-}
+    parseJSON = withText "Version" $ go . readP_to_S parseVersion . unpack
+      where
+        go [(v,[])] = return v
+        go (_ : xs) = go xs
+        go _        = fail $ "could not parse Version"
+
+instance ToJSON (Proxy a) where
+    toJSON _ = Null
+    {-# INLINE toJSON #-}
+
+    toEncoding _ = Encoding E.null_
+    {-# INLINE toEncoding #-}
+
+instance FromJSON (Proxy a) where
+    {-# INLINE parseJSON #-}
+    parseJSON Null = pure Proxy
+    parseJSON v    = typeMismatch "Proxy" v
+
+instance ToJSON b => ToJSON (Tagged a b) where
+    toJSON (Tagged x) = toJSON x
+    {-# INLINE toJSON #-}
+
+    toEncoding (Tagged x) = toEncoding x
+    {-# INLINE toEncoding #-}
+
+instance FromJSON b => FromJSON (Tagged a b) where
+    {-# INLINE parseJSON #-}
+    parseJSON = fmap Tagged . parseJSON
+
+instance ToJSON a => ToJSON (Const a b) where
+    toJSON (Const x) = toJSON x
+    {-# INLINE toJSON #-}
+
+    toEncoding (Const x) = toEncoding x
+    {-# INLINE toEncoding #-}
+
+instance FromJSON a => FromJSON (Const a b) where
+    {-# INLINE parseJSON #-}
+    parseJSON = fmap Const . parseJSON
+
 -- | @withObject expected f value@ applies @f@ to the 'Object' when @value@ is an @Object@
 --   and fails using @'typeMismatch' expected@ otherwise.
 withObject :: String -> (Object -> Parser a) -> Value -> Parser a
@@ -1523,7 +1651,7 @@ ifromJSON = iparse parseJSON
 --
 -- This accessor is appropriate if the key and value /must/ be present
 -- in an object for it to be valid.  If the key and value are
--- optional, use '(.:?)' instead.
+-- optional, use '.:?' instead.
 (.:) :: (FromJSON a) => Object -> Text -> Parser a
 obj .: key = case H.lookup key obj of
                Nothing -> fail $ "key " ++ show key ++ " not present"
@@ -1539,15 +1667,26 @@ obj .: key = case H.lookup key obj of
 --
 -- This accessor is most useful if the key and value can be absent
 -- from an object without affecting its validity.  If the key and
--- value are mandatory, use '(.:)' instead.
+-- value are mandatory, use '.:' instead.
 (.:?) :: (FromJSON a) => Object -> Text -> Parser (Maybe a)
 obj .:? key = case H.lookup key obj of
+               Nothing -> pure Nothing
+               Just v  -> modifyFailure addKeyName
+                        $ parseJSON v <?> Key key
+  where
+    addKeyName = (("failed to parse field " <> unpack key <> ": ") <>)
+{-# INLINE (.:?) #-}
+
+-- | Like '.:?', but the resulting parser will fail,
+-- if the key is present but is 'Null'.
+(.:!) :: (FromJSON a) => Object -> Text -> Parser (Maybe a)
+obj .:! key = case H.lookup key obj of
                Nothing -> pure Nothing
                Just v  -> modifyFailure addKeyName
                         $ Just <$> parseJSON v <?> Key key
   where
     addKeyName = (("failed to parse field " <> unpack key <> ": ") <>)
-{-# INLINE (.:?) #-}
+{-# INLINE (.:!) #-}
 
 -- | Helper for use in combination with '.:?' to provide default
 -- values for optional JSON object fields.
@@ -1555,7 +1694,7 @@ obj .:? key = case H.lookup key obj of
 -- This combinator is most useful if the key and value can be absent
 -- from an object without affecting its validity and we know a default
 -- value to assign in that case.  If the key and value are mandatory,
--- use '(.:)' instead.
+-- use '.:' instead.
 --
 -- Example usage:
 --
