@@ -1,11 +1,19 @@
-{-# LANGUAGE DeriveDataTypeable, GeneralizedNewtypeDeriving, Rank2Types,
+{-# LANGUAGE CPP, DeriveDataTypeable, GeneralizedNewtypeDeriving, Rank2Types,
     RecordWildCards #-}
+#if __GLASGOW_HASKELL__ >= 800
+-- a) THQ works on cross-compilers and unregisterised GHCs
+-- b) may make compilation faster as no dynamic loading is ever needed (not sure about this)
+-- c) removes one hindrance to have code inferred as SafeHaskell safe
+{-# LANGUAGE TemplateHaskellQuotes #-}
+#else
+{-# LANGUAGE TemplateHaskell #-}
+#endif
 
 -- |
 -- Module:      Data.Aeson.Types.Internal
--- Copyright:   (c) 2011-2015 Bryan O'Sullivan
+-- Copyright:   (c) 2011-2016 Bryan O'Sullivan
 --              (c) 2011 MailRank, Inc.
--- License:     Apache
+-- License:     BSD3
 -- Maintainer:  Bryan O'Sullivan <bos@serpentine.com>
 -- Stability:   experimental
 -- Portability: portable
@@ -17,6 +25,7 @@ module Data.Aeson.Types.Internal
     -- * Core JSON types
       Value(..)
     , Encoding(..)
+    , unsafeToEncoding
     , Series(..)
     , Array
     , emptyArray, isEmptyArray
@@ -53,26 +62,39 @@ module Data.Aeson.Types.Internal
     , DotNetTime(..)
     ) where
 
+import Control.Arrow (first)
 import Control.Applicative
 import Control.DeepSeq (NFData(..))
 import Control.Monad (MonadPlus(..), ap)
+import qualified Control.Monad.Fail as Fail
 import Data.ByteString.Builder (Builder, char7, toLazyByteString)
-import Data.Char (isLower, isUpper, toLower)
+import Data.Char (isLower, isUpper, toLower, isAlpha, isAlphaNum)
 import Data.Data (Data)
-import Data.Foldable (Foldable(..))
+import Data.Foldable (foldl')
 import Data.HashMap.Strict (HashMap)
 import Data.Hashable (Hashable(..))
-import Data.Monoid (Monoid(..), (<>))
+import Data.Semigroup (Semigroup((<>)))
 import Data.Scientific (Scientific)
 import Data.String (IsString(..))
 import Data.Text (Text, pack, unpack)
 import Data.Time (UTCTime)
 import Data.Time.Format (FormatTime)
-import Data.Traversable (Traversable(..))
 import Data.Typeable (Typeable)
 import Data.Vector (Vector)
 import qualified Data.HashMap.Strict as H
+import qualified Data.Scientific as S
 import qualified Data.Vector as V
+import qualified Language.Haskell.TH.Syntax as TH
+
+#if !MIN_VERSION_base(4,8,0)
+import Data.Foldable (Foldable(..))
+import Data.Monoid (Monoid(..))
+import Data.Traversable (Traversable(..))
+#endif
+
+#if !MIN_VERSION_unordered_containers(0,2,6)
+import Data.List (sort)
+#endif
 
 -- | Elements of a JSON path used to describe the location of an
 -- error.
@@ -118,35 +140,43 @@ instance Functor Result where
     {-# INLINE fmap #-}
 
 instance Monad IResult where
-    return = ISuccess
+    return = pure
     {-# INLINE return #-}
 
     ISuccess a      >>= k = k a
     IError path err >>= _ = IError path err
     {-# INLINE (>>=) #-}
 
+    fail = Fail.fail
+    {-# INLINE fail #-}
+
+instance Fail.MonadFail IResult where
     fail err = IError [] err
     {-# INLINE fail #-}
 
 instance Monad Result where
-    return = Success
+    return = pure
     {-# INLINE return #-}
 
     Success a >>= k = k a
     Error err >>= _ = Error err
     {-# INLINE (>>=) #-}
 
+    fail = Fail.fail
+    {-# INLINE fail #-}
+
+instance Fail.MonadFail Result where
     fail err = Error err
     {-# INLINE fail #-}
 
 instance Applicative IResult where
-    pure  = return
+    pure  = ISuccess
     {-# INLINE pure #-}
     (<*>) = ap
     {-# INLINE (<*>) #-}
 
 instance Applicative Result where
-    pure  = return
+    pure  = Success
     {-# INLINE pure #-}
     (<*>) = ap
     {-# INLINE (<*>) #-}
@@ -177,16 +207,24 @@ instance Alternative Result where
     (<|>) = mplus
     {-# INLINE (<|>) #-}
 
+instance Semigroup (IResult a) where
+    (<>) = mplus
+    {-# INLINE (<>) #-}
+
 instance Monoid (IResult a) where
     mempty  = fail "mempty"
     {-# INLINE mempty #-}
-    mappend = mplus
+    mappend = (<>)
     {-# INLINE mappend #-}
 
+instance Semigroup (Result a) where
+    (<>) = mplus
+    {-# INLINE (<>) #-}
+
 instance Monoid (Result a) where
     mempty  = fail "mempty"
     {-# INLINE mempty #-}
-    mappend = mplus
+    mappend = (<>)
     {-# INLINE mappend #-}
 
 instance Foldable IResult where
@@ -235,8 +273,12 @@ instance Monad Parser where
     m >>= g = Parser $ \path kf ks -> let ks' a = runParser (g a) path kf ks
                                        in runParser m path kf ks'
     {-# INLINE (>>=) #-}
-    return a = Parser $ \_path _kf ks -> ks a
+    return = pure
     {-# INLINE return #-}
+    fail = Fail.fail
+    {-# INLINE fail #-}
+
+instance Fail.MonadFail Parser where
     fail msg = Parser $ \path kf _ks -> kf (reverse path) msg
     {-# INLINE fail #-}
 
@@ -246,7 +288,7 @@ instance Functor Parser where
     {-# INLINE fmap #-}
 
 instance Applicative Parser where
-    pure  = return
+    pure a = Parser $ \_path _kf ks -> ks a
     {-# INLINE pure #-}
     (<*>) = apP
     {-# INLINE (<*>) #-}
@@ -264,10 +306,14 @@ instance MonadPlus Parser where
                                          in runParser a path kf' ks
     {-# INLINE mplus #-}
 
+instance Semigroup (Parser a) where
+    (<>) = mplus
+    {-# INLINE (<>) #-}
+
 instance Monoid (Parser a) where
     mempty  = fail "mempty"
     {-# INLINE mempty #-}
-    mappend = mplus
+    mappend = (<>)
     {-# INLINE mappend #-}
 
 apP :: Parser (a -> b) -> Parser a -> Parser b
@@ -296,7 +342,14 @@ data Value = Object !Object
 newtype Encoding = Encoding {
       fromEncoding :: Builder
       -- ^ Acquire the underlying bytestring builder.
-    } deriving (Monoid)
+    } deriving (Semigroup,Monoid)
+
+-- | Make Encoding from Builder.
+--
+-- Use with care! You have to make sure that the passed Builder
+-- is a valid JSON Encoding!
+unsafeToEncoding :: Builder -> Encoding
+unsafeToEncoding = Encoding
 
 instance Show Encoding where
     show (Encoding e) = show (toLazyByteString e)
@@ -308,21 +361,29 @@ instance Ord Encoding where
     compare (Encoding a) (Encoding b) =
       compare (toLazyByteString a) (toLazyByteString b)
 
--- | A series of values that, when encoded, should be separated by commas.
+-- | A series of values that, when encoded, should be separated by
+-- commas. Since 0.11.0.0, the '.=' operator is overloaded to create
+-- either @(Text, Value)@ or 'Series'. You can use Series when
+-- encoding directly to a bytestring builder as in the following
+-- example:
+--
+-- > toEncoding (Person name age) = pairs ("name" .= name <> "age" .= age)
 data Series = Empty
             | Value Encoding
             deriving (Typeable)
 
-instance Monoid Series where
-    mempty              = Empty
-
-    mappend Empty a     = a
-    mappend (Value a) b =
+instance Semigroup Series where
+    Empty   <> a = a
+    Value a <> b =
         Value $
         a <> case b of
                Empty   -> mempty
                Value c -> Encoding (char7 ',') <> c
 
+instance Monoid Series where
+    mempty  = Empty
+    mappend = (<>)
+
 -- | A newtype wrapper for 'UTCTime' that uses the same non-standard
 -- serialization format as Microsoft .NET, whose
 -- <https://msdn.microsoft.com/en-us/library/system.datetime(v=vs.110).aspx System.DateTime>
@@ -338,7 +399,7 @@ newtype DotNetTime = DotNetTime {
 
 instance NFData Value where
     rnf (Object o) = rnf o
-    rnf (Array a)  = V.foldl' (\x y -> rnf y `seq` x) () a
+    rnf (Array a)  = foldl' (\x y -> rnf y `seq` x) () a
     rnf (String s) = rnf s
     rnf (Number n) = rnf n
     rnf (Bool b)   = rnf b
@@ -349,9 +410,15 @@ instance IsString Value where
     {-# INLINE fromString #-}
 
 hashValue :: Int -> Value -> Int
-hashValue s (Object o)   = H.foldl' hashWithSalt
-                              (s `hashWithSalt` (0::Int)) o
-hashValue s (Array a)    = V.foldl' hashWithSalt
+#if MIN_VERSION_unordered_containers(0,2,6)
+hashValue s (Object o)   = s `hashWithSalt` (0::Int) `hashWithSalt` o
+#else
+hashValue s (Object o)   = foldl' hashWithSalt
+                              (s `hashWithSalt` (0::Int)) assocHashesSorted
+  where
+    assocHashesSorted = sort [hash k `hashWithSalt` v | (k, v) <- H.toList o]
+#endif
+hashValue s (Array a)    = foldl' hashWithSalt
                               (s `hashWithSalt` (1::Int)) a
 hashValue s (String str) = s `hashWithSalt` (2::Int) `hashWithSalt` str
 hashValue s (Number n)   = s `hashWithSalt` (3::Int) `hashWithSalt` n
@@ -361,6 +428,21 @@ hashValue s Null         = s `hashWithSalt` (5::Int)
 instance Hashable Value where
     hashWithSalt = hashValue
 
+-- @since 0.11.0.0
+instance TH.Lift Value where
+    lift Null = [| Null |]
+    lift (Bool b) = [| Bool b |]
+    lift (Number n) = [| Number (S.scientific c e) |]
+      where
+        c = S.coefficient n
+        e = S.base10Exponent n
+    lift (String t) = [| String (pack s) |]
+      where s = unpack t
+    lift (Array a) = [| Array (V.fromList a') |]
+      where a' = V.toList a
+    lift (Object o) = [| Object (H.fromList . map (first pack) $ o') |]
+      where o' = map (first unpack) . H.toList $ o
+
 -- | The empty array.
 emptyArray :: Value
 emptyArray = Array V.empty
@@ -402,9 +484,28 @@ parseEither m v = runParser (m v) [] onError Right
 formatError :: JSONPath -> String -> String
 formatError path msg = "Error in " ++ (format "$" path) ++ ": " ++ msg
   where
+    format :: String -> JSONPath -> String
     format pfx []                = pfx
     format pfx (Index idx:parts) = format (pfx ++ "[" ++ show idx ++ "]") parts
-    format pfx (Key key:parts)   = format (pfx ++ "." ++ unpack key) parts
+    format pfx (Key key:parts)   = format (pfx ++ formatKey key) parts
+
+    formatKey :: Text -> String
+    formatKey key
+       | isIdentifierKey strKey = "." ++ strKey
+       | otherwise              = "['" ++ escapeKey strKey ++ "']"
+      where strKey = unpack key
+
+    isIdentifierKey :: String -> Bool
+    isIdentifierKey []     = False
+    isIdentifierKey (x:xs) = isAlpha x && all isAlphaNum xs
+
+    escapeKey :: String -> String
+    escapeKey = concatMap escapeChar
+
+    escapeChar :: Char -> String
+    escapeChar '\'' = "\\'"
+    escapeChar '\\' = "\\\\"
+    escapeChar c    = [c]
 
 -- | A key\/value pair for an 'Object'.
 type Pair = (Text, Value)
@@ -417,19 +518,20 @@ object = Object . H.fromList
 
 -- | Add JSON Path context to a parser
 --
--- When parsing complex structure it helps to annotate (sub)parsers
--- with context so that if error occurs you can find it's location.
+-- When parsing a complex structure, it helps to annotate (sub)parsers
+-- with context, so that if an error occurs, you can find its location.
 --
 -- > withObject "Person" $ \o ->
 -- >   Person
 -- >     <$> o .: "name" <?> Key "name"
 -- >     <*> o .: "age"  <?> Key "age"
 --
--- (except for standard methods like '(.:)' already do that)
+-- (Standard methods like '(.:)' already do this.)
 --
--- After that in case of error you will get a JSON Path location of that error.
+-- With such annotations, if an error occurs, you will get a JSON Path
+-- location of that error.
 --
--- Since 0.9
+-- Since 0.10
 (<?>) :: Parser a -> JSONPathElement -> Parser a
 p <?> pathElem = Parser $ \path kf ks -> runParser p (pathElem:path) kf ks
 
@@ -520,6 +622,7 @@ data SumEncoding =
 -- , 'allNullaryToStringTag'   = True
 -- , 'omitNothingFields'       = False
 -- , 'sumEncoding'             = 'defaultTaggedObject'
+-- , 'unwrapUnaryRecords'      = False
 -- }
 -- @
 defaultOptions :: Options
