@@ -2,19 +2,26 @@
 
 module Properties where
 
+import Control.Applicative (Const)
 import Data.Aeson (eitherDecode)
-import Data.Aeson.Encode (encode, encodeToBuilder)
+import Data.Aeson.Encode (encode)
 import Data.Aeson.Internal (IResult(..), formatError, ifromJSON, iparse)
 import Data.Aeson.Parser (value)
 import Data.Aeson.Types
 import Data.ByteString.Builder (toLazyByteString)
 import Data.Int (Int8)
-import Data.Time (Day, LocalTime, NominalDiffTime, UTCTime, ZonedTime)
+import Data.List.NonEmpty (NonEmpty)
+import Data.Time (Day, LocalTime, NominalDiffTime, TimeOfDay, UTCTime,
+                  ZonedTime)
+import Data.Proxy (Proxy)
+import Data.Tagged (Tagged)
+import Data.Version (Version)
 import Encoders
 import Instances ()
+import Numeric.Natural (Natural)
 import Test.Framework (Test, testGroup)
 import Test.Framework.Providers.QuickCheck2 (testProperty)
-import Test.QuickCheck (Arbitrary(..), Property, (===))
+import Test.QuickCheck (Arbitrary(..), Property, (===), (.&&.), counterexample)
 import Types
 import qualified Data.Attoparsec.Lazy as L
 import qualified Data.ByteString.Lazy.Char8 as L
@@ -39,16 +46,31 @@ toParseJSON parsejson tojson x =
       IError path msg -> failure "parse" (formatError path msg) x
       ISuccess x'     -> x === x'
 
-roundTrip :: (FromJSON a, ToJSON a, Show a) =>
+roundTripEnc :: (FromJSON a, ToJSON a, Show a) =>
              (a -> a -> Property) -> a -> a -> Property
-roundTrip eq _ i =
-    case fmap ifromJSON . L.parse value . encode . toJSON $ i of
+roundTripEnc eq _ i =
+    case fmap ifromJSON . L.parse value . encode $ i of
       L.Done _ (ISuccess v)      -> v `eq` i
       L.Done _ (IError path err) -> failure "fromJSON" (formatError path err) i
       L.Fail _ _ err             -> failure "parse" err i
 
+roundTripNoEnc :: (FromJSON a, ToJSON a, Show a) =>
+             (a -> a -> Property) -> a -> a -> Property
+roundTripNoEnc eq _ i =
+    case ifromJSON . toJSON $ i of
+      (ISuccess v)      -> v `eq` i
+      (IError path err) -> failure "fromJSON" (formatError path err) i
+
 roundTripEq :: (Eq a, FromJSON a, ToJSON a, Show a) => a -> a -> Property
-roundTripEq x y = roundTrip (===) x y
+roundTripEq x y = roundTripEnc (===) x y .&&. roundTripNoEnc (===) x y
+
+infix 4 ==~
+(==~) :: (ApproxEq a, Show a) => a -> a -> Property
+x ==~ y =
+  counterexample (show x ++ " /= " ++ show y) (x =~ y)
+
+roundTripApproxEq :: (ApproxEq a, FromJSON a, ToJSON a, Show a) => a -> a -> Property
+roundTripApproxEq x y = roundTripEnc (==~) x y .&&. roundTripNoEnc (==~) x y
 
 toFromJSON :: (Arbitrary a, Eq a, FromJSON a, ToJSON a, Show a) => a -> Property
 toFromJSON x = case ifromJSON (toJSON x) of
@@ -63,16 +85,11 @@ modifyFailureProp orig added =
     result :: Result ()
     result = parse parser ()
 
--- | Perform a bit-for-bit comparison of two encoding methods.
+-- | Perform a structural comparison of the results of two encoding
+-- methods. Compares decoded values to account for HashMap-driven
+-- variation in JSON object key ordering.
 sameAs :: (a -> Value) -> (a -> Encoding) -> a -> Property
 sameAs toVal toEnc v =
-  toLazyByteString (encodeToBuilder (toVal v)) ===
-  toLazyByteString (fromEncoding (toEnc v))
-
--- | Behaves like 'sameAs', but compares decoded values to account for
--- HashMap-driven variation in JSON object key ordering.
-sameAsV :: (a -> Value) -> (a -> Encoding) -> a -> Property
-sameAsV toVal toEnc v =
   eitherDecode (toLazyByteString (fromEncoding (toEnc v))) === Right (toVal v)
 
 type P6 = Product6 Int Bool String (Approx Double) (Int, Approx Double) ()
@@ -115,16 +132,23 @@ tests = testGroup "properties" [
       testProperty "Bool" $ roundTripEq True
     , testProperty "Double" $ roundTripEq (1 :: Approx Double)
     , testProperty "Int" $ roundTripEq (1 :: Int)
+    , testProperty "NonEmpty Char" $ roundTripEq (undefined :: NonEmpty Char)
     , testProperty "Integer" $ roundTripEq (1 :: Integer)
     , testProperty "String" $ roundTripEq ("" :: String)
     , testProperty "Text" $ roundTripEq T.empty
     , testProperty "Foo" $ roundTripEq (undefined :: Foo)
     , testProperty "Day" $ roundTripEq (undefined :: Day)
-    , testProperty "DotNetTime" $ roundTripEq (undefined :: DotNetTime)
+    , testProperty "DotNetTime" $ roundTripApproxEq (undefined :: DotNetTime)
     , testProperty "LocalTime" $ roundTripEq (undefined :: LocalTime)
+    , testProperty "TimeOfDay" $ roundTripEq (undefined :: TimeOfDay)
     , testProperty "UTCTime" $ roundTripEq (undefined :: UTCTime)
     , testProperty "ZonedTime" $ roundTripEq (undefined :: ZonedTime)
     , testProperty "NominalDiffTime" $ roundTripEq (undefined :: NominalDiffTime)
+    , testProperty "Version" $ roundTripEq (undefined :: Version)
+    , testProperty "Natural" $ roundTripEq (undefined :: Natural)
+    , testProperty "Proxy" $ roundTripEq (undefined :: Proxy Int)
+    , testProperty "Tagged" $ roundTripEq (undefined :: Tagged Int Char)
+    , testProperty "Const" $ roundTripEq (undefined :: Const Int Char)
     , testGroup "ghcGenerics" [
         testProperty "OneConstructor" $ roundTripEq OneConstructor
       , testProperty "Product2" $ roundTripEq (undefined :: Product2 Int Bool)
@@ -174,6 +198,14 @@ tests = testGroup "properties" [
               , testProperty "ObjectWithSingleField" (toParseJSON thApproxParseJSONDefault thApproxToJSONDefault)
             ]
           ]
+        , testGroup "GADT" [
+            testProperty "string"                (isString                . thGADTToJSONUnwrap)
+          , testProperty "ObjectWithSingleField" (isObjectWithSingleField . thGADTToJSONDefault)
+          , testGroup "roundTrip" [
+                testProperty "string"                (toParseJSON thGADTParseJSONUnwrap  thGADTToJSONUnwrap)
+              , testProperty "ObjectWithSingleField" (toParseJSON thGADTParseJSONDefault thGADTToJSONDefault)
+            ]
+          ]
         ]
       ]
     , testGroup "toEncoding" [
@@ -191,11 +223,11 @@ tests = testGroup "properties" [
       , testProperty "ApproxDefault" $
         thApproxToJSONDefault `sameAs` thApproxToEncodingDefault
       , testProperty "SomeType2ElemArray" $
-        thSomeTypeToJSON2ElemArray `sameAsV` thSomeTypeToEncoding2ElemArray
+        thSomeTypeToJSON2ElemArray `sameAs` thSomeTypeToEncoding2ElemArray
       , testProperty "SomeTypeTaggedObject" $
-        thSomeTypeToJSONTaggedObject `sameAsV` thSomeTypeToEncodingTaggedObject
+        thSomeTypeToJSONTaggedObject `sameAs` thSomeTypeToEncodingTaggedObject
       , testProperty "SomeTypeObjectWithSingleField" $
-        thSomeTypeToJSONObjectWithSingleField `sameAsV`
+        thSomeTypeToJSONObjectWithSingleField `sameAs`
         thSomeTypeToEncodingObjectWithSingleField
       ]
     ]
